﻿using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.Instances;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using System;

namespace HeuristicLab.Problems.PTSP {
  [Item("Estimated Probabilistic Traveling Salesman Problem", "Represents an estimated Probabilistic Traveling Salesman Problem.")]
  [Creatable("Problems")]
  [StorableClass]
  public sealed class EstimatedProbabilisticTravelingSalesmanProblem : ProbabilisticTravelingSalesmanProblem, IStorableContent,
  IProblemInstanceConsumer<TSPData> {

    [StorableConstructor]
    private EstimatedProbabilisticTravelingSalesmanProblem(bool deserializing) : base(deserializing) { }
    private EstimatedProbabilisticTravelingSalesmanProblem(EstimatedProbabilisticTravelingSalesmanProblem original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new EstimatedProbabilisticTravelingSalesmanProblem(this, cloner);
    }

    public override double Evaluate(Individual individual, IRandom random) {
      Permutation p = individual.Permutation();
      // Estimation-based evaluation
      Random r = new Random();
      double estimatedSum = 0;
      for (int i = 0; i < SampleSize.Value; i++) {
        int singleRealization = -1;
        for (int j = 0; j < p.Length - 1; j++) {
          if (ProbabilityMatrix[0, p[j]] > r.NextDouble()) {
            if (singleRealization != -1) {
              estimatedSum += DistanceMatrix[singleRealization, p[j]];
            }
            singleRealization = p[j];
          }
        }
      }
      return estimatedSum / SampleSize.Value;
    }

    public EstimatedProbabilisticTravelingSalesmanProblem() {
      Parameters.Add(new ValueParameter<IntValue>("SampleSize", "Size of the sample for the estimation-based evaluation"));
      Operators.Add(new PTSPInversionMovePathEvaluator());
    }

    public override void Load(TSPData data) {
      base.Load(data);
      // For now uses sample size of 20 but should use Student's t-test
      SampleSize = new IntValue(20);
    }
  }
}